/*const shape = [2,3,3,3,1];

const weights = [
  [[1,1,0]  ,[1,1,0]  ,[1,1,0]  ],//[[inputSize][inputSize]..layerSize]
  [[1,1,1,0],[1,1,1,0],[1,1,1,0]],
  [[1,1,1,0],[1,1,1,0],[1,1,1,0]],
  [[1,1,1,0]                    ]
];*/

function randomWeights(shape, w = []) {
  const weights = [];
  for (var l=1; l<shape.length; l++) {
    weights[l-1] = [];
    for (var i=0; i<shape[l]; i++) {
      weights[l-1][i] = [];
      for (var o=0; o<shape[l-1] + 1; o++) {
        weights[l-1][i][o] = (w[l-1] && w[l-1][i] && w[l-1][i][o]) || (2*Math.random()-1);
      }
    }
  }
  return weights;
}

function f(input, weights) {
  const output = [];
  output[-1] = input;
  for (var l=0; l<weights.length; l++) {
    output[l] = [];
    for (var i=0; i<weights[l].length; i++) {
      output[l][i] = 0;
      for (var o=0; o<output[l-1].length; o++) {
        output[l][i] += output[l-1][o] * weights[l][i][o];
      }
      output[l][i] += weights[l][i][output[l-1].length]; // Bias
      output[l][i] *= output[l][i] < 0 ? -0.001 : 1;
    }
  }
  return output;
}

const neuron = (a, w) => {
  // a: Activations
  // w: Weights for this neuron + bias
  // z: New activation
  // k: Previous activation index
  let z = 0;
  for (var k = 0; k < w.length - 1; k++) {
    z += a[k] * w[k];
  }
  return z + w[w.length - 1];
}

function train(inputs, targets, weights) {
  const inverse_gradiant = [];
  for (var i=0; i<inputs.length; i++) {
    var activations = f(inputs[i], weights);
    let deltaCost0_deltaActivation = 2 * activations[activations.length-1]
        .map((currentValue, index) => currentValue - targets[i][index])
        .reduce((previousValue, currentValue, currentIndex, array) => previousValue + currentValue);
    for (var layer_index = weights.length - 1; layer_index >= 0; layer_index--) {
      inverse_gradiant[layer_index] = inverse_gradiant[layer_index] || [];
      let deltaCost0_deltaPrevActivation = 0;
      for (var activation_index = 0; activation_index < weights[layer_index].length; activation_index++) {
        
        inverse_gradiant[layer_index][activation_index] = inverse_gradiant[layer_index][activation_index] || [];
        inverse_gradiant[layer_index][activation_index][weights[layer_index][activation_index].length - 1] =
        inverse_gradiant[layer_index][activation_index][weights[layer_index][activation_index].length - 1] || 0;
        const deltaZ_deltaPrevActivation = activations[layer_index-1].reduce((previousValue, currentValue, currentIndex, array) => previousValue + currentValue);
        let z = neuron(activations[layer_index - 1], weights[layer_index][activation_index]);
        const deltaActivation_deltaZ = z < 0 ? 0.01 : 1;
        deltaCost0_deltaPrevActivation += deltaZ_deltaPrevActivation * deltaActivation_deltaZ * deltaCost0_deltaActivation;
        
        for (var prevActivation_index = 0; prevActivation_index < weights[layer_index][activation_index].length - 1; prevActivation_index++) {
          inverse_gradiant[layer_index][activation_index][prevActivation_index] =
          inverse_gradiant[layer_index][activation_index][prevActivation_index] || 0;
          const deltaZ_deltaWeights = activations[layer_index-1][prevActivation_index];
          const deltaCost0_deltaWeights = deltaZ_deltaWeights * deltaActivation_deltaZ * deltaCost0_deltaActivation;
          inverse_gradiant[layer_index][activation_index][prevActivation_index] += deltaCost0_deltaWeights / inputs.length;
        }

        const deltaZ_deltaBias = 1;
        const deltaCost0_deltaBias = deltaZ_deltaBias * deltaActivation_deltaZ * deltaCost0_deltaActivation;
        inverse_gradiant[layer_index][activation_index][weights[layer_index][activation_index].length - 1] += deltaCost0_deltaBias / inputs.length;
      }
      deltaCost0_deltaActivation = deltaCost0_deltaPrevActivation;
    }
  }
  return inverse_gradiant;

}

function printWeights(weights) {
  console.log('Weights: [');
  for (var l=0; l<weights.length; l++) {
    text = '[';
    for (var i=0; i<weights[l].length; i++) {
      text += '[';
      for (var w=0; w<weights[l][i].length; w++) {
        text += Math.floor(weights[l][i][w]*100)/100 + ',';
      }
      text = text.slice(0,-1) + '],';
    }
    console.log(text + '],');
  }
  console.log(']');
}


const run = () => {

  printWeights(weights);

  /*
  Weights: [
  ai.js:78
  [[0.91,-1,1,-0.65],[-1,1,1,-0.82],[0.83,0.86,0.3,-0.63],],
  ai.js:88
  [[-1,1,-1,0.29],[-1,1,-1,-0.68],[0.69,-0.53,0.69,-0.09],[0.4,-0.39,0.31,0.66],],
  ai.js:88
  [[1,0.8],[-1,0.09],[0.87,0.66],[0.89,0.24],],
  ai.js:88
  [[-0.57],[0.55],],
  ai.js:88
  ]
  */
  var valid = true;
  for (var next_layer_index = 1, current_layer_index = 0; next_layer_index < weights.length; next_layer_index++, current_layer_index++) {
    for (var node_index = 0; node_index < weights[next_layer_index].length; node_index++) {
      valid &= weights[next_layer_index][node_index].length == weights[current_layer_index].length + 1;
    }
  }
  if (!valid) {
    console.error("weights is not valid. first layer must contain nodes of the same length as second layer");
  }

  console.log(f([0,0]).reverse()[0]);


  const X = [[0,0],[0,1],[1,0],[1,1]];
  const Y = [[0],[1],[1],[0]];

  train(X,Y, weights);
  printWeights(weights);
  train(X,Y, weights);
  printWeights(weights);
  var n = 1;
  while ((n = parseInt(prompt("Continue?", n))) > 0) {
    for (var i=0; i<n; i++) {
      train(X,Y, weights);
    }
    printWeights(weights);


    console.log('0,0:'+Math.floor(f([0,0], weights).reverse()[0]*1000)/1000);
    console.log('0,1:'+Math.floor(f([0,1], weights).reverse()[0]*1000)/1000);
    console.log('1,0:'+Math.floor(f([1,0], weights).reverse()[0]*1000)/1000);
    console.log('1,1:'+Math.floor(f([1,1], weights).reverse()[0]*1000)/1000);
  }
}

//run();